import copy
import pkg_resources

import dclab
import numpy as np
from PyQt5 import uic, QtCore, QtWidgets

from ...pipeline import Plot
from ...pipeline.plot import STATE_OPTIONS


COLORMAPS = STATE_OPTIONS["scatter"]["colormap"]


class PlotPanel(QtWidgets.QWidget):
    #: Emitted when a shapeout2.pipeline.Plot is to be changed
    plot_changed = QtCore.pyqtSignal(dict)
    #: Emitted when the pipeline is to be changed
    pipeline_changed = QtCore.pyqtSignal(dict)

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "ana_plot.ui")
        uic.loadUi(path_ui, self)

        # current Shape-Out 2 pipeline
        self._pipeline = None
        self._init_controls()
        self.update_content()

        # options for division
        self.comboBox_division.clear()
        self.comboBox_division.addItem("Merge all plots", "merge")
        self.comboBox_division.addItem("One plot per dataset", "each")
        self.comboBox_division.addItem("Scatter plots and joint contour plot",
                                       "multiscatter+contour")
        self.comboBox_division.setCurrentIndex(2)

        # signals
        self.toolButton_duplicate.clicked.connect(self.on_plot_duplicated)
        self.toolButton_remove.clicked.connect(self.on_plot_removed)
        self.pushButton_reset.clicked.connect(self.update_content)
        self.pushButton_apply.clicked.connect(self.write_plot)
        self.comboBox_plots.currentIndexChanged.connect(self.update_content)
        self.comboBox_marker_hue.currentIndexChanged.connect(
            self.on_hue_selected)
        self.comboBox_marker_feature.currentIndexChanged.connect(
            self.on_hue_selected)
        self.comboBox_axis_x.currentIndexChanged.connect(self.on_axis_changed)
        self.comboBox_axis_y.currentIndexChanged.connect(self.on_axis_changed)
        self.comboBox_scale_x.currentIndexChanged.connect(self.on_axis_changed)
        self.comboBox_scale_y.currentIndexChanged.connect(self.on_axis_changed)
        self.spinBox_column_count.valueChanged.connect(
            self.on_column_num_changed)
        self.widget_range_x.range_changed.connect(self.on_range_changed)
        self.widget_range_y.range_changed.connect(self.on_range_changed)

    def __getstate__(self):
        feats_srt = self.get_features()

        rx = self.widget_range_x.__getstate__()
        ry = self.widget_range_y.__getstate__()

        # hue min/max
        marker_hue = self.comboBox_marker_hue.currentData()
        if marker_hue == "kde":
            hmin = 0
            hmax = 1
        elif marker_hue == "feature":
            rstate = self.widget_range_feat.__getstate__()
            hmin = rstate["start"]
            hmax = rstate["end"]
        else:
            hmin = hmax = np.nan

        state = {
            "identifier": self.current_plot.identifier,
            "layout": {
                "column count": self.spinBox_column_count.value(),
                "division": self.comboBox_division.currentData(),
                "label plots": self.checkBox_label_plots.isChecked(),
                "name": self.lineEdit.text(),
                "size x": self.spinBox_size_x.value(),
                "size y": self.spinBox_size_y.value(),
            },
            "general": {
                "auto range": self.checkBox_auto_range.isChecked(),
                "axis x": feats_srt[self.comboBox_axis_x.currentIndex()],
                "axis y": feats_srt[self.comboBox_axis_y.currentIndex()],
                "isoelastics": self.checkBox_isoelastics.isChecked(),
                "kde": self.comboBox_kde.currentData(),
                "range x": [rx["start"], rx["end"]],
                "range y": [ry["start"], ry["end"]],
                "scale x": self.comboBox_scale_x.currentData(),
                "scale y": self.comboBox_scale_y.currentData(),
            },
            "scatter": {
                "colormap": self.comboBox_colormap.currentData(),
                "downsample": self.checkBox_downsample.isChecked(),
                "downsampling value": self.spinBox_downsample.value(),
                "enabled": self.groupBox_scatter.isChecked(),
                "hue feature": self.comboBox_marker_feature.currentData(),
                "hue max": hmax,
                "hue min": hmin,
                "marker alpha": self.spinBox_alpha.value() / 100,
                "marker hue": marker_hue,
                "marker size": self.doubleSpinBox_marker_size.value(),
                "show event count": self.checkBox_event_count.isChecked(),
            },
            "contour": {
                "enabled": self.groupBox_contour.isChecked(),
                "legend":  self.checkBox_legend.isChecked(),
                "line widths": [self.doubleSpinBox_lw_1.value(),
                                self.doubleSpinBox_lw_2.value(),
                                ],
                "line styles": [self.comboBox_ls_1.currentData(),
                                self.comboBox_ls_2.currentData(),
                                ],
                "percentiles": [self.doubleSpinBox_perc_1.value(),
                                self.doubleSpinBox_perc_2.value(),
                                ],
                "spacing x": self.doubleSpinBox_spacing_x.value(),
                "spacing y": self.doubleSpinBox_spacing_y.value(),
            }
        }
        return state

    def __setstate__(self, state):
        if self.current_plot.identifier != state["identifier"]:
            raise ValueError("Plot identifier mismatch!")
        feats_srt = self.get_features()
        toblock = [
            self.comboBox_axis_x,
            self.comboBox_axis_y,
            self.widget_range_x,
            self.widget_range_y,
        ]

        for b in toblock:
            b.blockSignals(True)

        # Layout
        lay = state["layout"]
        self.spinBox_column_count.setValue(lay["column count"])
        idx = self.comboBox_division.findData(lay["division"])
        self.comboBox_division.setCurrentIndex(idx)
        self.checkBox_label_plots.setChecked(lay["label plots"])
        self.lineEdit.setText(lay["name"])
        self.spinBox_size_x.setValue(lay["size x"])
        self.spinBox_size_y.setValue(lay["size y"])
        # General
        gen = state["general"]
        self.checkBox_auto_range.setChecked(gen["auto range"])
        self.comboBox_axis_x.setCurrentIndex(feats_srt.index(gen["axis x"]))
        self.comboBox_axis_y.setCurrentIndex(feats_srt.index(gen["axis y"]))
        self.checkBox_isoelastics.setChecked(gen["isoelastics"])
        kde_index = self.comboBox_kde.findData(gen["kde"])
        self.comboBox_kde.setCurrentIndex(kde_index)
        scx_index = self.comboBox_scale_x.findData(gen["scale x"])
        self.comboBox_scale_x.setCurrentIndex(scx_index)
        scy_index = self.comboBox_scale_y.findData(gen["scale y"])
        self.comboBox_scale_y.setCurrentIndex(scy_index)
        self._set_range_xy_state(axis_x=gen["axis x"],
                                 axis_y=gen["axis y"],
                                 range_x=gen["range x"],
                                 range_y=gen["range y"],
                                 )

        # Scatter
        sca = state["scatter"]
        self.checkBox_downsample.setChecked(sca["downsample"])
        self.spinBox_downsample.setValue(sca["downsampling value"])
        self.groupBox_scatter.setChecked(sca["enabled"])
        hue_index = self.comboBox_marker_hue.findData(sca["marker hue"])
        self.comboBox_marker_hue.setCurrentIndex(hue_index)
        self.doubleSpinBox_marker_size.setValue(sca["marker size"])
        if sca["hue feature"] in feats_srt:
            feat_index = feats_srt.index(sca["hue feature"])
        else:
            feat_index = 0  # feature not available in datasets
        self.comboBox_marker_feature.setCurrentIndex(feat_index)
        color_index = COLORMAPS.index(sca["colormap"])
        self.comboBox_colormap.setCurrentIndex(color_index)
        self.checkBox_event_count.setChecked(sca["show event count"])
        self.spinBox_alpha.setValue(int(sca["marker alpha"]*100))
        if sca["marker hue"] == "feature":
            self._set_range_feat_state(sca["hue feature"], sca["hue min"],
                                       sca["hue max"])

        # Contour
        con = state["contour"]
        self.groupBox_contour.setChecked(con["enabled"])
        self.checkBox_legend.setChecked(con["legend"])
        self.doubleSpinBox_perc_1.setValue(con["percentiles"][0])
        self.doubleSpinBox_perc_2.setValue(con["percentiles"][1])
        self.doubleSpinBox_lw_1.setValue(con["line widths"][0])
        self.doubleSpinBox_lw_2.setValue(con["line widths"][1])
        ls1_index = self.comboBox_ls_1.findData(con["line styles"][0])
        self.comboBox_ls_1.setCurrentIndex(ls1_index)
        ls2_index = self.comboBox_ls_2.findData(con["line styles"][1])
        self.comboBox_ls_2.setCurrentIndex(ls2_index)
        self._set_contour_spacing(spacing_x=con["spacing x"],
                                  spacing_y=con["spacing y"])
        for b in toblock:
            b.blockSignals(False)

    def _init_controls(self):
        """All controls that are not subject to change"""
        # KDE
        kde_names = STATE_OPTIONS["general"]["kde"]
        self.comboBox_kde.clear()
        for kn in kde_names:
            self.comboBox_kde.addItem(kn.capitalize(), kn)
        # Scales
        scales = STATE_OPTIONS["general"]["scale x"]
        self.comboBox_scale_x.clear()
        self.comboBox_scale_y.clear()
        for sc in scales:
            if sc == "log":
                vc = "logarithmic"
            else:
                vc = sc
            self.comboBox_scale_x.addItem(vc, sc)
            self.comboBox_scale_y.addItem(vc, sc)
        # Marker hue
        hues = STATE_OPTIONS["scatter"]["marker hue"]
        self.comboBox_marker_hue.clear()
        for hue in hues:
            if hue == "kde":
                huev = "KDE"
            else:
                huev = hue.capitalize()
            self.comboBox_marker_hue.addItem(huev, hue)
        self.comboBox_colormap.clear()
        for c in COLORMAPS:
            self.comboBox_colormap.addItem(c, c)
        # Contour line styles
        lstyles = STATE_OPTIONS["contour"]["line styles"][0]
        self.comboBox_ls_1.clear()
        self.comboBox_ls_2.clear()
        for ls in lstyles:
            self.comboBox_ls_1.addItem(ls, ls)
            self.comboBox_ls_2.addItem(ls, ls)
        # range controls
        for rc in [self.widget_range_x, self.widget_range_y,
                   self.widget_range_feat]:
            rc.setLabel("")
            rc.setCheckable(False)
        # hide feature label range selection
        self.widget_range_feat.hide()

    def _set_range_feat_state(self, feat, fmin=None, fmax=None):
        """Set a proper state for the feature hue range control"""
        if len(self.pipeline.slots) == 0:
            self.setEnabled(False)
            # do nothing
            return
        else:
            self.setEnabled(True)
        if feat is not None:
            lim = self.pipeline.get_min_max(
                feat=feat, plot_id=self.current_plot.identifier)
            if not (np.isinf(lim[0]) or np.isinf(lim[1])):
                self.widget_range_feat.setLimits(vmin=lim[0], vmax=lim[1])
                if fmin is None:
                    fmin = lim[0]
                if fmax is None:
                    fmax = lim[1]
                self.widget_range_feat.__setstate__({"active": True,
                                                     "start": fmin,
                                                     "end": fmax,
                                                     })

    def _set_range_xy_state(self, axis_x=None, range_x=None,
                            axis_y=None, range_y=None):
        """Set a proper state for the x/y range controls"""
        if len(self.pipeline.slots) == 0:
            self.setEnabled(False)
            # do nothing
            return
        else:
            self.setEnabled(True)

        plot_id = self.current_plot.identifier

        for axis, rang, rc in zip([axis_x, axis_y],
                                  [range_x, range_y],
                                  [self.widget_range_x, self.widget_range_y],
                                  ):
            if axis is not None:
                lim = self.pipeline.get_min_max(feat=axis, plot_id=plot_id)
                if not (np.isinf(lim[0]) or np.isinf(lim[1])):
                    rc.setLimits(vmin=lim[0],
                                 vmax=lim[1])
                    if rang is None or rang[0] == rang[1]:
                        # default range is limits + 5% margin
                        rang = self.pipeline.get_min_max(feat=axis,
                                                         plot_id=plot_id,
                                                         margin=0.05)
                    rc.__setstate__({"active": True,
                                     "start": rang[0],
                                     "end": rang[1],
                                     })

    def _set_contour_spacing(self, spacing_x=None, spacing_y=None):
        """Set the contour spacing in the spin boxes

        - sets spinbox limits first
        - sets number of digits
        - sets step
        - sets value in the end
        """
        for spacing, spinBox in zip([spacing_x, spacing_y],
                                    [self.doubleSpinBox_spacing_x,
                                     self.doubleSpinBox_spacing_y]):
            if spacing is None:
                continue
            else:
                if spacing >= 1:
                    dec = 2
                else:
                    dec = -np.int(np.log10(spacing)) + 3
                spinBox.setDecimals(dec)
                spinBox.setMinimum(10**-dec)
                spinBox.setMaximum(max(10*spacing, 10))
                spinBox.setSingleStep(10**(-dec + 1))
                spinBox.setValue(spacing)

    def _set_contour_spacing_auto(self, axis_x=None, axis_y=None):
        """automatically set the contour spacing

        - uses :func:`dclab.kde_methods.bin_width_percentile`
        - uses _set_contour_spacing
        """
        if len(self.pipeline.slots) == 0:
            self.setEnabled(False)
            # do nothing
            return
        else:
            self.setEnabled(True)
        dslist, _ = self.pipeline.get_plot_datasets(
            self.current_plot.identifier)
        if dslist:
            spacings_xy = []
            for axis, scaleCombo in zip([axis_x, axis_y],
                                        [self.comboBox_scale_x,
                                         self.comboBox_scale_y]):
                if axis is None:
                    # nothing to do
                    spacings_xy.append(None)
                else:
                    # determine good approximation
                    spacings = []
                    for ds in dslist:
                        spa = ds.get_kde_spacing(
                            a=ds[axis],
                            feat=axis,
                            scale=scaleCombo.currentData(),
                            method=dclab.kde_methods.bin_width_percentile,
                        )
                        spacings.append(spa)
                    spacings_xy.append(np.min(spacings))
            spacing_x, spacing_y = spacings_xy
            # sets the limits before setting the value
            self._set_contour_spacing(spacing_x=spacing_x,
                                      spacing_y=spacing_y)

    @property
    def current_plot(self):
        if self.plot_ids:
            plot_index = self.comboBox_plots.currentIndex()
            plot_id = self.plot_ids[plot_index]
            plot = Plot.get_instances()[plot_id]
        else:
            plot = None
        return plot

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def plot_ids(self):
        """List of plot identifiers"""
        if self.pipeline is not None:
            ids = [plot.identifier for plot in self.pipeline.plots]
        else:
            ids = []
        return ids

    @property
    def plot_names(self):
        """List of plot names"""
        if self.pipeline is not None:
            ids = [plot.name for plot in self.pipeline.plots]
        else:
            ids = []
        return ids

    def get_features(self):
        """Wrapper around pipeline with default features if empty"""
        feats_srt = self.pipeline.get_features(
            scalar=True, label_sort=True, plot_id=self.current_plot.identifier)
        if len(feats_srt) == 0:
            # fallback (nothing in the pipeline)
            features = dclab.dfn.scalar_feature_names
            labs = [dclab.dfn.get_feature_label(f) for f in features]
            lf = sorted(zip(labs, features))
            feats_srt = [it[1] for it in lf]
        return feats_srt

    def on_axis_changed(self):
        gen = self.__getstate__()["general"]
        if self.sender() == self.comboBox_axis_x:
            self._set_range_xy_state(axis_x=gen["axis x"])
            self._set_contour_spacing_auto(axis_x=gen["axis x"])
        elif self.sender() == self.comboBox_axis_y:
            self._set_range_xy_state(axis_y=gen["axis y"])
            self._set_contour_spacing_auto(axis_y=gen["axis y"])
        elif self.sender() == self.comboBox_scale_x:
            self._set_contour_spacing_auto(axis_x=gen["axis x"])
        elif self.sender() == self.comboBox_scale_y:
            self._set_contour_spacing_auto(axis_y=gen["axis y"])

    def on_column_num_changed(self):
        """The user changed the number of columns

        - increase/decrease self.spinBox_size_x by 150pt
        - increase/decrease self.spinBox_size_y by 150pt if
          the row count changes as well
        """
        # old parameters
        state = self.current_plot.__getstate__()
        plot_id = state["identifier"]
        plot_index = self.pipeline.plot_ids.index(plot_id)
        old_size_x = state["layout"]["size x"]
        old_size_y = state["layout"]["size y"]
        old_ncol, old_nrow = self.pipeline.get_plot_col_row_count(plot_id)
        # new parameters
        new_pipeline_state = self.pipeline.__getstate__()
        new_pipeline_state["plots"][plot_index] = self.__getstate__()
        new_ncol, new_nrow = self.pipeline.get_plot_col_row_count(
            plot_id, new_pipeline_state)
        # size x (minimum of 400)
        new_size_x = max(400, old_size_x + 200*(new_ncol - old_ncol))
        self.spinBox_size_x.setValue(new_size_x)
        # size y
        new_size_y = max(400, old_size_y + 200*(new_nrow - old_nrow))
        self.spinBox_size_y.setValue(new_size_y)

    def on_hue_selected(self):
        """Show/hide options for feature-based hue selection"""
        selection = self.comboBox_marker_hue.currentData()
        # hide everything
        self.comboBox_marker_feature.hide()
        self.widget_dataset_alpha.hide()
        self.comboBox_colormap.hide()
        self.label_colormap.hide()
        self.widget_range_feat.hide()
        # Only show feature selection if needed
        if selection == "feature":
            self.comboBox_marker_feature.show()
            self.comboBox_colormap.show()
            self.label_colormap.show()
            self.widget_range_feat.show()
            # set the range
            self._set_range_feat_state(
                feat=self.comboBox_marker_feature.currentData())
        elif selection == "kde":
            self.comboBox_colormap.show()
            self.label_colormap.show()
        elif selection in ["dataset", "none"]:
            self.widget_dataset_alpha.show()
        else:
            raise ValueError("Unknown selection: '{}'".format(selection))

    def on_plot_duplicated(self):
        # determine the new filter state
        plot_state = self.__getstate__()
        new_state = copy.deepcopy(plot_state)
        new_plot = Plot()
        new_state["identifier"] = new_plot.identifier
        new_state["layout"]["name"] = new_plot.name
        new_plot.__setstate__(new_state)
        # determine the filter position
        pos = self.pipeline.plot_ids.index(plot_state["identifier"])
        self.pipeline.add_plot(new_plot, index=pos+1)
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    def on_plot_removed(self):
        plot_state = self.__getstate__()
        self.pipeline.remove_plot(plot_state["identifier"])
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    def on_range_changed(self):
        """User changed x/y range -> disable auto range checkbox"""
        self.checkBox_auto_range.setChecked(False)

    def show_plot(self, plot_id):
        self.update_content(plot_index=self.plot_ids.index(plot_id))

    def set_pipeline(self, pipeline):
        self._pipeline = pipeline

    def update_content(self, event=None, plot_index=None):
        if self.plot_ids:
            self.setEnabled(True)
            # update combobox
            self.comboBox_plots.blockSignals(True)
            # this also updates the combobox
            if plot_index is None:
                plot_index = self.comboBox_plots.currentIndex()
                if plot_index > len(self.plot_ids) - 1 or plot_index < 0:
                    plot_index = len(self.plot_ids) - 1
            self.comboBox_plots.clear()
            self.comboBox_plots.addItems(self.plot_names)
            self.comboBox_plots.setCurrentIndex(plot_index)
            self.comboBox_plots.blockSignals(False)
            # set choices for all comboboxes that deal with features
            for cb in [self.comboBox_axis_x,
                       self.comboBox_axis_y,
                       self.comboBox_marker_feature]:
                # get the features currently available
                feats_srt = self.get_features()
                cb.blockSignals(True)
                # remember previous selection if possible
                if cb.count:
                    # remember current selection
                    curfeat = cb.currentData()
                    if curfeat not in feats_srt:
                        curfeat = None
                else:
                    curfeat = None
                # repopulate
                cb.clear()
                for feat in feats_srt:
                    cb.addItem(dclab.dfn.get_feature_label(feat), feat)
                if curfeat is not None:
                    # write back current selection
                    curidx = feats_srt.index(curfeat)
                    cb.setCurrentIndex(curidx)
                cb.blockSignals(False)
            # populate content
            plot = Plot.get_plot(identifier=self.plot_ids[plot_index])
            state = plot.__getstate__()
            self.__setstate__(state)
        else:
            self.setEnabled(False)

    def write_plot(self):
        """Update the shapeout2.pipeline.Plot instance"""
        # get current index
        plot_index = self.comboBox_plots.currentIndex()
        plot = Plot.get_plot(identifier=self.plot_ids[plot_index])
        plot_state = self.__getstate__()
        plot.__setstate__(plot_state)
        self.update_content()  # update plot selection combobox
        self.plot_changed.emit(plot_state)



# flake8: noqa: F401
from .dlg_slot_reorder import DlgSlotReorder
from .ana_filter import FilterPanel
from .ana_meta import MetaPanel
from .ana_plot import PlotPanel
from .ana_slot import SlotPanel
from .ana_view import AnalysisView



import pathlib
import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets
import pyqtgraph.exporters as pge

from ..pipeline_plot import PipelinePlot
from ..widgets import show_wait_cursor

from ...util import get_valid_filename


EXPORTERS = {
    "png": ["rendered image (*.png)", pge.ImageExporter],
    "svg": ["vector graphics (*.svg)", pge.SVGExporter],
}


class ExportPlot(QtWidgets.QDialog):
    def __init__(self, parent, pipeline, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.export", "e2plot.ui")
        uic.loadUi(path_ui, self)
        # set pipeline
        self.pipeline = pipeline
        # populate combobox plots
        self.comboBox_plot.clear()
        self.comboBox_plot.addItem("All plots", "all")
        for plot in pipeline.plots:
            self.comboBox_plot.addItem(plot.name, plot.identifier)
        # populate combobox format
        self.comboBox_fmt.clear()
        for key in EXPORTERS:
            self.comboBox_fmt.addItem(EXPORTERS[key][0], key)
        # Signals
        self.comboBox_fmt.currentIndexChanged.connect(self.on_format)

    def done(self, r):
        if r:
            self.export_plots()
        super(ExportPlot, self).done(r)

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def export_plots(self):
        # show dialog
        fmt = self.comboBox_fmt.currentData()
        # keys are plot identifiers, values are paths
        fnames = {}
        if self.comboBox_plot.currentData() == "all":
            path = QtWidgets.QFileDialog.getExistingDirectory(self,
                                                              'Output Folder')
            if path:
                for ii, plot in enumerate(self.pipeline.plots):
                    fn = "SO-plot_{}_{}.{}".format(ii, plot.name, fmt)
                    # remove bad characters from file name
                    fn = get_valid_filename(fn)
                    fnames[plot.identifier] = pathlib.Path(path) / fn
        else:
            pp, _ = QtWidgets.QFileDialog.getSaveFileName(
                self, 'Plot export file name', '',
                self.comboBox_fmt.currentText())
            if pp:
                fnames[self.comboBox_plot.currentData()] = pathlib.Path(pp)

        # get PipelinePlot instance
        for plot_id in fnames:
            pipl = PipelinePlot.instances[plot_id]
            exp = EXPORTERS[fmt][1](pipl.plot_layout.centralWidget)
            if fmt == "png":
                dpi = self.spinBox_dpi.value()
                exp.params["width"] = int(exp.params["width"] / 72 * dpi)
                exp.params["antialias"] = self.checkBox_aa.isChecked()
            pout = str(fnames[plot_id])
            if not pout.endswith(fmt):
                pout += "."+fmt
            exp.export(pout)

    def on_format(self):
        if self.comboBox_fmt.currentData() == "png":
            self.widget_png.show()
        else:
            self.widget_png.hide()



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets


class DlgSlotReorder(QtWidgets.QDialog):
    pipeline_changed = QtCore.pyqtSignal(dict)

    def __init__(self, pipeline, *args, **kwargs):
        super(DlgSlotReorder, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "dlg_slot_reorder.ui")
        uic.loadUi(path_ui, self)

        self.pipeline = pipeline
        for ii, slot in enumerate(pipeline.slots):
            self.listWidget.addItem("{}: {}".format(ii, slot.name))

        self.toolButton_down.clicked.connect(self.on_move_item)
        self.toolButton_up.clicked.connect(self.on_move_item)
        btn_ok = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        btn_ok.clicked.connect(self.on_ok)

    @QtCore.pyqtSlot()
    def on_ok(self):
        """Apply the changes made in the UI and update the pipeline"""
        # get order
        indices = []
        for row in range(self.listWidget.count()):
            item = self.listWidget.item(row)
            text = item.text()
            idx = int(text.split(":", 1)[0])
            indices.append(idx)
        # reorder pipeline and send pipeline_changed signal
        self.pipeline.reorder_slots(indices)
        state = self.pipeline.__getstate__()
        self.pipeline_changed.emit(state)

    @QtCore.pyqtSlot()
    def on_move_item(self):
        """Move currently selected item one row up or down"""
        row = self.listWidget.currentRow()
        if row == -1:
            return
        item = self.listWidget.takeItem(row)

        if self.sender() == self.toolButton_down:
            new_row = row + 1
        else:
            new_row = row - 1

        self.listWidget.insertItem(new_row, item)
        self.listWidget.setCurrentRow(new_row)



import pathlib
import pkg_resources
import time

from PyQt5 import uic, QtCore, QtWidgets

import dclab

from ..widgets import show_wait_cursor

from ...util import get_valid_filename
from ..._version import version


class ExportData(QtWidgets.QDialog):
    def __init__(self, parent, pipeline, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.export", "e2data.ui")
        uic.loadUi(path_ui, self)
        # Get output path
        self.on_browse()
        # set pipeline
        self.pipeline = pipeline
        # update list widget
        self.on_radio()
        # Signals
        self.pushButton_path.clicked.connect(self.on_browse)
        self.toolButton_all.clicked.connect(self.on_select_all)
        self.toolButton_none.clicked.connect(self.on_select_none)
        self.radioButton_fcs.clicked.connect(self.on_radio)
        self.radioButton_rtdc.clicked.connect(self.on_radio)
        self.radioButton_tsv.clicked.connect(self.on_radio)

    @property
    def file_format(self):
        if self.radioButton_fcs.isChecked():
            return "fcs"
        elif self.radioButton_rtdc.isChecked():
            return "rtdc"
        else:
            return "tsv"

    def done(self, r):
        if r:
            self.export_data()
        super(ExportData, self).done(r)

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def export_data(self):
        """Export data to the desired file format"""
        out = pathlib.Path(self.path)
        # get features
        features = []
        for ii in range(self.listWidget.count()):
            if self.listWidget.item(ii).checkState() == 2:
                features.append(self.features[ii])
        pend = len(self.pipeline.slots)
        prog = QtWidgets.QProgressDialog("Exporting...", "Abort", 1,
                                         pend, self)
        prog.setWindowTitle("Data Export")
        prog.setWindowModality(QtCore.Qt.WindowModal)
        prog.setMinimumDuration(0)
        time.sleep(0.01)
        prog.setValue(0)
        QtWidgets.QApplication.processEvents()
        for slot_index in range(len(self.pipeline.slots)):
            slot = self.pipeline.slots[slot_index]
            if slot.slot_used:  # only export slots "used" (#15)
                ds = self.pipeline.get_dataset(slot_index)
                fn = "SO2-export_{}_{}.{}".format(slot_index,
                                                  slot.name,
                                                  self.file_format)
                # remove bad characters from file name
                fn = get_valid_filename(fn)
                path = out / fn
                if self.file_format == "rtdc":
                    ds.export.hdf5(path=path,
                                   features=features,
                                   override=True)
                elif self.file_format == "fcs":
                    ds.export.fcs(path=path,
                                  features=features,
                                  meta_data={"Shape-Out version": version},
                                  override=True)
                else:
                    ds.export.tsv(path=path,
                                  features=features,
                                  meta_data={"Shape-Out version": version},
                                  override=True)
            if prog.wasCanceled():
                break
            prog.setValue(slot_index + 1)
            QtWidgets.QApplication.processEvents()
        prog.setValue(pend)

    def on_browse(self):
        out = QtWidgets.QFileDialog.getExistingDirectory(self,
                                                         'Export directory')
        if out:
            self.path = out
            self.lineEdit_path.setText(self.path)
        else:
            self.path = None

    def on_radio(self):
        self.update_feature_list()

    def on_select_all(self):
        for ii in range(self.listWidget.count()):
            wid = self.listWidget.item(ii)
            wid.setCheckState(2)

    def on_select_none(self):
        for ii in range(self.listWidget.count()):
            wid = self.listWidget.item(ii)
            wid.setCheckState(0)

    def update_feature_list(self, scalar=False):
        if self.file_format == "rtdc":
            self.features = self.pipeline.get_features(union=True,
                                                       label_sort=True)
            # do not allow exporting event index, since it will be
            # re-enumerated in any case.
            self.features.remove("index")
        else:
            self.features = self.pipeline.get_features(scalar=True,
                                                       union=True,
                                                       label_sort=True)

        self.listWidget.clear()
        for feat in self.features:
            wid = QtWidgets.QListWidgetItem(dclab.dfn.get_feature_label(feat))
            wid.setCheckState(2)
            self.listWidget.addItem(wid)



# flake8: noqa: F401
from .bg_thread import run_async
from .colorbar_widget import ColorBarWidget
from .double_spin_box_nan import DoubleSpinBoxNan
from .key_value_table_widget import KeyValueTableWidget
from .mdi_subwindow_wo_close import MDISubWindowWOButtons
from .qrangeslider import QRangeSlider
from .rangecontrol import RangeControl
from .simple_image_view import SimpleImageView
from .simple_plot_widget import (SimplePlotItem,
                                 SimplePlotWidget,
                                 SimpleViewBox)
from .wait_cursor import show_wait_cursor, ShowWaitCursor



"""Test bulk action for emodulus computation"""
import pathlib
import tempfile

import dclab
import h5py
import numpy as np
import pytest

from shapeout2.gui.main import ShapeOut2
from shapeout2.gui import bulk
from shapeout2 import session


datapath = pathlib.Path(__file__).parent / "data"


def make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23],
                 chip_region="channel"):
    # create a fake dataset
    path = datapath / "calibration_beads_47.rtdc"
    ds = dclab.new_dataset(path)
    tmp = tempfile.mktemp(".rtdc", prefix="example_")
    ds.export.hdf5(tmp, features=["deform", "area_um", "bright_avg"])
    with h5py.File(tmp, mode="a") as h5:
        h5["events/temp"] = np.linspace(temp_range[0], temp_range[1], len(ds))
        if medium is None:
            h5.attrs.pop("setup:medium")
        else:
            h5.attrs["setup:medium"] = medium
        h5.attrs["setup:temperature"] = temp
        h5.attrs["setup:chip region"] = chip_region
    return pathlib.Path(tmp)


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    session.clear_session()
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    session.clear_session()


def test_manual_basic(qtbot):
    """Most simple test"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])
    mw.add_dataslot(paths=[path])

    # sanity check (no emodulus should be available)
    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" not in ds
        assert ds.config["setup"]["medium"] == "CellCarrierB"

    # create bulk action dialog manually
    dlg = bulk.BulkActionEmodulus(mw, pipeline=mw.pipeline)
    dlg.comboBox_temp.setCurrentIndex(dlg.comboBox_temp.findData("manual"))
    dlg.doubleSpinBox_temp.setValue(29.5)
    dlg.on_ok()

    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" in ds
        assert ds.config["setup"]["medium"] == "CellCarrierB"
        assert ds.config["calculation"]["emodulus model"] == "elastic sphere"
        assert ds.config["calculation"]["emodulus medium"] == "CellCarrierB"
        assert ds.config["calculation"]["emodulus temperature"] == 29.5
        assert "emodulus viscosity" not in ds.config["calculation"]


def test_manual_wrong_medium(qtbot):
    """Deliberately set wrong medium"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])
    mw.add_dataslot(paths=[path])

    # create bulk action dialog manually
    dlg = bulk.BulkActionEmodulus(mw, pipeline=mw.pipeline)
    dlg.comboBox_temp.setCurrentIndex(dlg.comboBox_temp.findData("manual"))
    dlg.doubleSpinBox_temp.setValue(29.5)
    # Set medium to "water". This should not change the emodulus medium.
    dlg.comboBox_medium.setCurrentIndex(dlg.comboBox_medium.findData("water"))

    dlg.on_ok()

    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" in ds
        assert ds.config["setup"]["medium"] == "CellCarrierB"
        assert ds.config["calculation"]["emodulus model"] == "elastic sphere"
        assert ds.config["calculation"]["emodulus medium"] == "CellCarrierB"
        assert ds.config["calculation"]["emodulus temperature"] == 29.5
        assert "emodulus viscosity" not in ds.config["calculation"]


def test_temperature_feature(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add custom dataslot
    path = make_dataset(medium="CellCarrier", temp=22.5, temp_range=[22, 23])
    mw.add_dataslot(paths=[path])
    mw.add_dataslot(paths=[path])

    # create bulk action dialog manually
    dlg = bulk.BulkActionEmodulus(mw, pipeline=mw.pipeline)
    dlg.comboBox_temp.setCurrentIndex(dlg.comboBox_temp.findData("feature"))
    dlg.on_ok()

    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" in ds
        assert ds.config["setup"]["medium"] == "CellCarrier"
        assert ds.config["calculation"]["emodulus model"] == "elastic sphere"
        assert ds.config["calculation"]["emodulus medium"] == "CellCarrier"
        assert "emodulus temperature" not in ds.config["calculation"]
        assert "emodulus viscosity" not in ds.config["calculation"]


def test_viscosity(qtbot):
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add custom dataslot
    path = make_dataset(medium="other")
    mw.add_dataslot(paths=[path])
    mw.add_dataslot(paths=[path])

    # create bulk action dialog manually
    dlg = bulk.BulkActionEmodulus(mw, pipeline=mw.pipeline)
    dlg.comboBox_medium.setCurrentIndex(dlg.comboBox_medium.findData("other"))
    dlg.doubleSpinBox_visc.setValue(1.0)
    dlg.on_ok()

    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" in ds
        assert ds.config["setup"]["medium"] == "other"
        assert ds.config["calculation"]["emodulus model"] == "elastic sphere"
        assert "emodulus medium" not in ds.config["calculation"]
        assert "emodulus temperature" not in ds.config["calculation"]
        assert ds.config["calculation"]["emodulus viscosity"] == 1.0


def test_wrong_medium_viscosity(qtbot):
    """Deliberately set wrong visosity"""
    mw = ShapeOut2()
    qtbot.addWidget(mw)

    # add a dataslot
    path = datapath / "calibration_beads_47.rtdc"
    mw.add_dataslot(paths=[path])
    mw.add_dataslot(paths=[path])

    # create bulk action dialog manually
    dlg = bulk.BulkActionEmodulus(mw, pipeline=mw.pipeline)
    dlg.comboBox_medium.setCurrentIndex(dlg.comboBox_medium.findData("other"))
    dlg.doubleSpinBox_visc.setValue(1.0)  # random number

    dlg.on_ok()

    for slot in mw.pipeline.slots:
        ds = slot.get_dataset()
        assert "emodulus" not in ds, "because medium is fixed"
        assert ds.config["setup"]["medium"] == "CellCarrierB"
        assert ds.config["calculation"]["emodulus model"] == "elastic sphere"
        assert ds.config["calculation"]["emodulus medium"] == "CellCarrierB"
        assert "emodulus temperature" not in ds.config["calculation"]
        assert "emodulus viscosity" not in ds.config["calculation"]



import hashlib
import numbers
import pathlib

import h5py
import numpy as np


def hashobj(obj):
    """Compute md5 hex-hash of a Python object"""
    return hashlib.md5(obj2bytes(obj)).hexdigest()


def obj2bytes(obj):
    """String representation of an object for hashing"""
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, pathlib.Path):
        return obj2bytes(str(obj))
    elif isinstance(obj, (bool, numbers.Number)):
        return str(obj).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(obj.items()))
    elif hasattr(obj, "identifier"):
        return obj2bytes(obj.identifier)
    elif isinstance(obj, h5py.Dataset):
        return obj2bytes(obj[0])
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))


def get_valid_filename(value):
    """
    Return the given string converted to a string that can be used
    for a clean filename.
    """
    ret = ""

    valid = "abcdefghijklmnopqrstuvwxyz" \
            + "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
            + "0123456789" \
            + "._-()"
    replace = {
        " ": "_",
        "[": "(",
        "]": ")",
        "µ": "u",
    }

    for ch in value:
        if ch in valid:
            ret += ch
        elif ch in replace:
            ret += replace[ch]
        else:
            ret += "?"

    ret = ret.strip(".")
    return ret



"""Keep record of all filters used"""
import copy

import dclab

from ..util import hashobj


class Filter(object):
    """Handles filters in a pipeline"""
    _instance_counter = 0
    _instances = {}

    def __init__(self, identifier=None, name=None):
        Filter._instance_counter += 1
        if identifier is None:
            identifier = "Filter_{}".format(Filter._instance_counter)
            while identifier in Filter._instances:
                Filter._instance_counter += 1
                identifier = "Filter_{}".format(Filter._instance_counter)

        if name is None:
            name = identifier
        #: unique identifier of the filter
        self.identifier = identifier
        #: user-defined name of the filter
        self.name = name
        #: general filtering arguments are directly passed to
        #: :class:`dclab.filter.Filter`
        self.general = {"remove invalid events": False,  # removes nan/inf
                        "enable filters": True,  # whether to use the filter
                        }
        #: filter for limiting number of events
        self.limit_events = [False, 5000]
        #: box filters with features as keys; each item is a
        #: dictionary with the keys "start", "end", "active"
        self.boxdict = {}
        #: polygon filter list; each item is an identifier of
        #: :class:`dclab.PolygonFilter`
        self.polylist = []
        if identifier in Filter._instances:
            raise ValueError("Filter with identifier "
                             + "'{}' already exists!".format(identifier))
        Filter._instances[identifier] = self

    def __getstate__(self):
        state = {
            "filter used": self.filter_used,
            "identifier": self.identifier,
            "limit events bool": self.limit_events[0],
            "limit events num": self.limit_events[1],
            "name": self.name,
            "remove invalid events": self.general["remove invalid events"],
            "box filters": self.boxdict,
            "polygon filters": sorted(self.polylist),
        }
        return copy.deepcopy(state)

    def __repr__(self):
        repre = "<Pipeline Filter '{}' at {}>".format(self.identifier,
                                                      hex(id(self)))
        return repre

    def __setstate__(self, state):
        state = copy.deepcopy(state)
        if self.identifier != state["identifier"]:
            raise ValueError("Identifier mismatch: '{}' vs. '{}'".format(
                self.identifier, state["identifier"]))
        self.general["enable filters"] = state["filter used"]
        self.limit_events = [state["limit events bool"],
                             state["limit events num"]]
        self.name = state["name"]
        self.general["remove invalid events"] = state["remove invalid events"]
        self.boxdict = state["box filters"]
        self.polylist = state["polygon filters"]

    @staticmethod
    def get_filter(identifier):
        """Get the filter with the given identifier.

        Notes
        -----
        Creates the filter if it does not exist.
        """
        if identifier in Filter._instances:
            f = Filter._instances[identifier]
        else:
            f = Filter(identifier=identifier)
        return f

    @staticmethod
    def get_instances():
        return Filter._instances

    @property
    def filter_used(self):
        return self.general["enable filters"]

    @filter_used.setter
    def filter_used(self, b):
        self.general["enable filters"] = b

    @property
    def hash(self):
        """Return the hash of the filter"""
        return hashobj(self.__getstate__())

    def add_box_filter(self, feature, start, end, active=True):
        """Add a box filter"""
        if not dclab.dfn.scalar_feature_exists(feature):
            raise ValueError("Unknown scalar feature: {}".format(feature))
        self.boxdict[feature] = {
            "start": start,
            "end": end,
            "active": active}

    def apply_to_dataset(self, dataset):
        """Convenience function to apply this filter to a dataset

        Parameters
        ----------
        dataset: :class:`dclab.RTDCBase`
            Input dataset

        Returns
        -------
        filtered_dataset: RTDCBase
            Either the input dataset, or a hierarchy child.

        Notes
        -----

        """
        self.update_dataset(dataset)
        dataset.apply_filter()

    def update_dataset(self, dataset):
        """Update the filtering configuration of a dataset

        Notes
        -----
        Due to the design of the filtering pipeline, it is not
        possible to use manual filters. If any are set, they
        are removed from the filter.
        """
        # remove all previous filters
        dataset.reset_filter()
        cfgfilt = dataset.config["filtering"]

        # set general options
        cfgfilt.update(self.general)
        cfgfilt["limit events"] = self.limit_events[0] * self.limit_events[1]
        # set box filters
        for feat in self.boxdict:
            if self.boxdict[feat]["active"]:
                cfgfilt["{} min".format(feat)] = self.boxdict[feat]["start"]
                cfgfilt["{} max".format(feat)] = self.boxdict[feat]["end"]

        # set polygon filters
        for pid in self.polylist:
            dataset.polygon_filter_add(pid)



import pathlib
import pkg_resources

import dclab
import numpy as np
from PyQt5 import uic, QtCore, QtWidgets
import pyqtgraph as pg
from scipy.ndimage import binary_erosion

from ..compute.comp_stats import STAT_METHODS
from .. import idiom
from ..widgets import show_wait_cursor


class QuickView(QtWidgets.QWidget):
    polygon_filter_created = QtCore.pyqtSignal()
    polygon_filter_modified = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.quick_view", "qv_main.ui")
        uic.loadUi(path_ui, self)
        path_css = pkg_resources.resource_filename(
            "shapeout2.gui.quick_view", "qv_style.css")
        stylesheet = pathlib.Path(path_css).read_text()
        self.groupBox_image.setStyleSheet(stylesheet)
        self.groupBox_trace.setStyleSheet(stylesheet)

        self.setWindowTitle("Quick View")

        self._set_initial_ui()

        # Set scale options (with data)
        for cb in [self.comboBox_xscale, self.comboBox_yscale]:
            cb.clear()
            cb.addItem("linear", "linear")
            cb.addItem("logarithmic", "log")

        # settings button
        self.toolButton_event.toggled.connect(self.on_tool)
        self.toolButton_poly.toggled.connect(self.on_tool)
        self.toolButton_settings.toggled.connect(self.on_tool)

        # polygon filter signals
        self.label_poly_create.hide()
        self.label_poly_modify.hide()
        self.pushButton_poly_save.hide()
        self.pushButton_poly_cancel.hide()
        self.pushButton_poly_create.clicked.connect(self.on_poly_create)
        self.pushButton_poly_save.clicked.connect(self.on_poly_done)
        self.pushButton_poly_cancel.clicked.connect(self.on_poly_done)
        self.comboBox_poly.currentIndexChanged.connect(self.on_poly_modify)
        self.update_polygon_panel()

        # event changed signal
        self.widget_scatter.scatter.sigClicked.connect(
            self.on_event_scatter_clicked)
        self.widget_scatter.update_hover_pos.connect(
            self.on_event_scatter_hover)
        self.spinBox_event.valueChanged.connect(self.on_event_scatter_spin)
        self.checkBox_image_contour.stateChanged.connect(
            self.on_event_scatter_update)
        self.checkBox_image_contrast.stateChanged.connect(
            self.on_event_scatter_update)
        self.checkBox_image_zoom.stateChanged.connect(
            self.on_event_scatter_update)
        self.checkBox_trace_raw.stateChanged.connect(
            self.on_event_scatter_update)
        self.checkBox_trace_legend.stateChanged.connect(
            self.on_event_scatter_update)
        self.checkBox_trace_zoom.stateChanged.connect(
            self.on_event_scatter_update)
        self.tabWidget_event.currentChanged.connect(
            self.on_event_scatter_update)

        # value changed signals for plot
        self.signal_widgets = [self.checkBox_downsample,
                               self.spinBox_downsample,
                               self.comboBox_x,
                               self.comboBox_y,
                               self.comboBox_xscale,
                               self.comboBox_yscale,
                               self.checkBox_isoelastics,
                               ]
        for w in self.signal_widgets:
            if hasattr(w, "currentIndexChanged"):
                w.currentIndexChanged.connect(self.plot)
            elif hasattr(w, "stateChanged"):
                w.stateChanged.connect(self.plot)
            elif hasattr(w, "valueChanged"):
                w.valueChanged.connect(self.plot)
        # copy statistics to clipboard
        self.toolButton_stats2clipboard.clicked.connect(
            self.on_stats2clipboard)

        # Set individual plots
        kw0 = dict(x=np.arange(10), y=np.arange(10))
        self.trace_plots = {
            "fl1_raw": pg.PlotDataItem(pen="#6EA068", **kw0),  # green
            "fl2_raw": pg.PlotDataItem(pen="#8E7A45", **kw0),  # orange
            "fl3_raw": pg.PlotDataItem(pen="#8F4D48", **kw0),  # red
            "fl1_median": pg.PlotDataItem(pen="#15BF00", **kw0),  # green
            "fl2_median": pg.PlotDataItem(pen="#BF8A00", **kw0),  # orange
            "fl3_median": pg.PlotDataItem(pen="#BF0C00", **kw0),  # red
        }
        for key in self.trace_plots:
            self.graphicsView_trace.addItem(self.trace_plots[key])
            self.trace_plots[key].hide()

        self.graphicsView_trace.plotItem.setLabels(
            left="Fluorescence [a.u.]", bottom="Event time [µs]")
        self.legend_trace = self.graphicsView_trace.addLegend(
            offset=(-.01, +.01))

        #: default parameters for the event image
        self.imkw = dict(autoLevels=False,
                         levels=(0, 255),
                         )

        # set initial empty dataset
        self._rtdc_ds = None
        self.slot = None

    def __getstate__(self):
        plot = {
            "downsampling": self.checkBox_downsample.isChecked(),
            "downsampling value": self.spinBox_downsample.value(),
            "axis x": self.comboBox_x.currentData(),
            "axis y": self.comboBox_y.currentData(),
            "scale x": self.comboBox_xscale.currentData(),
            "scale y": self.comboBox_yscale.currentData(),
            "isoelastics": self.checkBox_isoelastics.isChecked(),
        }
        event = {
            "index": self.spinBox_event.value(),
            "image auto contrast": self.checkBox_image_contrast.isChecked(),
            "image contour": self.checkBox_image_contour.isChecked(),
            "image zoom": self.checkBox_image_zoom.isChecked(),
            "trace legend": self.checkBox_trace_legend.isChecked(),
            "trace raw": self.checkBox_trace_raw.isChecked(),
            "trace zoom": self.checkBox_trace_zoom.isChecked(),
        }
        state = {
            "plot": plot,
            "event": event,
        }
        return state

    def __setstate__(self, state):
        plot = state["plot"]
        for tb in self.signal_widgets:
            tb.blockSignals(True)
        # downsampling
        self.checkBox_downsample.setChecked(plot["downsampling"])
        self.spinBox_downsample.setValue(plot["downsampling value"])
        # axes labels
        self.update_feature_choices()
        idx = self.comboBox_x.findData(plot["axis x"])
        self.comboBox_x.setCurrentIndex(idx)
        idy = self.comboBox_y.findData(plot["axis y"])
        self.comboBox_y.setCurrentIndex(idy)
        # scaling
        idxs = self.comboBox_xscale.findData(plot["scale x"])
        self.comboBox_xscale.setCurrentIndex(idxs)
        idys = self.comboBox_yscale.findData(plot["scale y"])
        self.comboBox_yscale.setCurrentIndex(idys)
        # isoelastics
        self.checkBox_isoelastics.setChecked(plot["isoelastics"])
        for tb in self.signal_widgets:
            tb.blockSignals(False)
        if "event" in state:
            event = state["event"]
            self.checkBox_image_contrast.setChecked(
                event["image auto contrast"])
            self.checkBox_image_contour.setChecked(event["image contour"])
            self.checkBox_image_zoom.setChecked(event["image zoom"])
            self.spinBox_event.setValue(event["index"])
            self.checkBox_trace_raw.setChecked(event["trace raw"])
            self.checkBox_trace_legend.setChecked(event["trace legend"])

    def _check_file_open(self, rtdc_ds):
        """Check whether a dataset is still open"""
        if isinstance(rtdc_ds, dclab.rtdc_dataset.RTDC_HDF5):
            if rtdc_ds._h5:
                # the file is open
                isopen = True
            else:
                isopen = False
        elif isinstance(rtdc_ds, dclab.rtdc_dataset.RTDC_Hierarchy):
            isopen = self._check_file_open(rtdc_ds.hparent)
        else:
            # DCOR
            isopen = True
        return isopen

    def _set_initial_ui(self):
        # Initially, only show the info about how QuickView works
        self.widget_tool.setEnabled(False)
        self.widget_scatter.hide()
        # Hide settings/events by default
        self.widget_event.hide()
        self.widget_settings.hide()
        self.widget_poly.hide()
        # show the how-to label
        self.label_howto.show()
        # hide the no events label
        self.label_noevents.hide()

    @property
    def rtdc_ds(self):
        """Dataset to plot; set to None initially and if the file is closed"""
        if self._rtdc_ds is not None:
            if not self._check_file_open(self._rtdc_ds):
                self._rtdc_ds = None
        # now check again
        if self._rtdc_ds is None:
            self._set_initial_ui()
        return self._rtdc_ds

    @rtdc_ds.setter
    def rtdc_ds(self, rtdc_ds):
        self._rtdc_ds = rtdc_ds

    def get_event_image(self, ds, event):
        state = self.__getstate__()
        imkw = self.imkw.copy()
        cellimg = ds["image"][event]
        # automatic contrast
        if state["event"]["image auto contrast"]:
            vmin, vmax = cellimg.min(), cellimg.max()
            cellimg = (cellimg - vmin) / (vmax - vmin) * 255
        # convert to RGB
        cellimg = cellimg.reshape(
            cellimg.shape[0], cellimg.shape[1], 1)
        cellimg = np.repeat(cellimg, 3, axis=2)
        # convert to int
        cellimg = np.require(cellimg, np.uint8, 'C')

        # Only load contour data if there is an image column.
        # We don't know how big the images should be so we
        # might run into trouble displaying random contours.
        if "mask" in ds and len(ds["mask"]) > event:
            mask = ds["mask"][event]
            if state["event"]["image contour"]:
                # compute contour image from mask
                cont = mask ^ binary_erosion(mask)
                # set red contour pixel values in original image
                cellimg[cont, 0] = int(255*.7)
                cellimg[cont, 1] = 0
                cellimg[cont, 2] = 0
            if state["event"]["image zoom"]:
                xv, yv = np.where(mask)
                idminx = xv.min() - 5
                idminy = yv.min() - 5
                idmaxx = xv.max() + 5
                idmaxy = yv.max() + 5
                idminx = idminx if idminx >= 0 else 0
                idminy = idminy if idminy >= 0 else 0
                shx, shy = mask.shape
                idmaxx = idmaxx if idmaxx < shx else shx
                idmaxy = idmaxy if idmaxy < shy else shy
                cellimg = cellimg[idminx:idmaxx, idminy:idmaxy]
        return cellimg, imkw

    def get_statistics(self):
        if self.rtdc_ds is not None:
            features = [self.comboBox_x.currentData(),
                        self.comboBox_y.currentData()]
            h, v = dclab.statistics.get_statistics(ds=self.rtdc_ds,
                                                   features=features,
                                                   methods=STAT_METHODS)
            return h, v
        else:
            return None, None

    def on_event_scatter_clicked(self, plot, point):
        """User clicked on scatter plot

        Parameters
        ----------
        plot: pg.PlotItem
            Active plot
        point: QPoint
            Selected point (determined by scatter plot widget)
        """
        # `self.on_tool` (`self.toolButton_event`) takes care of this:
        # self.widget_scatter.select.show()
        if not self.toolButton_event.isChecked():
            # emulate mouse toggle
            self.toolButton_event.setChecked(True)
            self.toolButton_event.toggled.emit(True)
        if self.widget_scatter.events_plotted is not None:
            # plotted events
            plotted = self.widget_scatter.events_plotted
            # get corrected index
            ds_idx = np.where(plotted)[0][point.index()]
            self.show_event(ds_idx)

    def on_event_scatter_hover(self, pos):
        """Update the image view in the polygon widget """
        if self.rtdc_ds is not None and self.toolButton_poly.isChecked():
            # plotted events
            plotted = self.widget_scatter.events_plotted
            spos = self.widget_scatter.scatter.mapFromView(pos)
            point = self.widget_scatter.scatter.pointAt(spos)
            # get corrected index
            event = np.where(plotted)[0][point.index()]
            if "image" in self.rtdc_ds:
                cellimg, imkw = self.get_event_image(self.rtdc_ds, event)
                self.imageView_image_poly.setImage(cellimg, **imkw)
                self.imageView_image_poly.show()
            else:
                self.imageView_image_poly.hide()

    def on_event_scatter_spin(self, event):
        """Sping control for event selection changed"""
        self.show_event(event - 1)

    def on_event_scatter_update(self):
        """Just update the event shown"""
        event = self.spinBox_event.value()
        self.show_event(event - 1)

    def on_poly_create(self):
        """User wants to create a polygon filter"""
        self.pushButton_poly_create.setEnabled(False)
        if not self.toolButton_poly.isChecked():
            # emulate mouse toggle
            self.toolButton_poly.setChecked(True)
            self.toolButton_poly.toggled.emit(True)
        self.comboBox_poly.setEnabled(False)
        self.groupBox_poly.setEnabled(True)
        self.label_poly_create.setVisible(True)
        self.pushButton_poly_save.setVisible(True)
        self.pushButton_poly_cancel.setVisible(True)
        # defaults
        self.lineEdit_poly.setText("Polygon Filter {}".format(
            dclab.PolygonFilter._instance_counter + 1))
        self.checkBox_poly.setChecked(False)
        self.widget_scatter.activate_poly_mode()
        # trigger resize and redraw
        mdiwin = self.parent()
        mdiwin.adjustSize()
        mdiwin.update()
        self.update()

    def on_poly_done(self):
        """User is done creating or modifying a polygon filter"""
        self.pushButton_poly_create.setEnabled(True)
        self.label_poly_create.setVisible(False)
        self.label_poly_modify.setVisible(False)
        self.pushButton_poly_save.setVisible(False)
        self.pushButton_poly_cancel.setVisible(False)
        if self.sender() == self.pushButton_poly_save:
            # save the polygon filter
            points = self.widget_scatter.get_poly_points()
            name = self.lineEdit_poly.text()
            inverted = self.checkBox_poly.isChecked()
            axes = self.widget_scatter.xax, self.widget_scatter.yax
            # determine whether to create a new polygon filter or whether
            # to update an existing one.
            idp = self.comboBox_poly.currentData()
            if idp is None:
                dclab.PolygonFilter(axes=axes, points=points, name=name,
                                    inverted=inverted)
                mode = "create"
            else:
                pf = dclab.PolygonFilter.get_instance_from_id(idp)
                pf.name = name
                pf.inverted = inverted
                pf.points = points
                mode = "modify"
        else:
            mode = "none"
        # remove the PolyLineRoi
        self.widget_scatter.activate_scatter_mode()
        self.update_polygon_panel()
        if mode == "create":
            self.polygon_filter_created.emit()
        elif mode == "modify":
            self.polygon_filter_modified.emit()

    def on_poly_modify(self):
        """User wants to modify a polygon filter"""
        self.pushButton_poly_create.setEnabled(False)
        self.comboBox_poly.setEnabled(False)
        self.groupBox_poly.setEnabled(True)
        self.label_poly_modify.setVisible(True)
        self.pushButton_poly_save.setVisible(True)
        self.pushButton_poly_cancel.setVisible(True)
        # get the polygon filter id
        idp = self.comboBox_poly.currentData()
        pf = dclab.PolygonFilter.get_instance_from_id(idp)
        # set UI information
        self.lineEdit_poly.setText(pf.name)
        self.checkBox_poly.setChecked(pf.inverted)
        # set axes
        state = self.__getstate__()
        state["plot"]["axis x"] = pf.axes[0]
        state["plot"]["axis y"] = pf.axes[1]
        self.__setstate__(state)
        self.plot()
        # add ROI
        self.widget_scatter.activate_poly_mode(pf.points)

    def on_stats2clipboard(self):
        """Copy the statistics as tsv data to the clipboard"""
        h, v = self.get_statistics()
        if h is not None:
            # assemble tsv data
            tsv = ""
            for hi, vi in zip(h, v):
                tsv += "{}\t{:.7g}\n".format(hi, vi)
            QtWidgets.qApp.clipboard().setText(tsv)

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def on_tool(self, collapse=False):
        """Show and hide tools when the user selected a tool button"""
        toblock = [self.toolButton_event,
                   self.toolButton_poly,
                   self.toolButton_settings,
                   ]
        for b in toblock:
            b.blockSignals(True)
        # show extra data
        show_event = False
        show_poly = False
        show_settings = False
        sender = self.sender()
        if sender == self.toolButton_event:
            show_event = self.toolButton_event.isChecked()
        elif sender == self.toolButton_poly:
            show_poly = self.toolButton_poly.isChecked()
        elif sender == self.toolButton_settings:
            show_settings = self.toolButton_settings.isChecked()
        elif collapse:
            # show nothing
            pass
        else:
            # keep everything as-is but update the sizes
            show_event = self.widget_event.isVisible()
            show_settings = self.widget_settings.isVisible()
            show_poly = self.widget_poly.isVisible()

        # toolbutton checked
        self.toolButton_event.setChecked(show_event)
        self.toolButton_poly.setChecked(show_poly)
        self.toolButton_settings.setChecked(show_settings)

        # widget visibility
        self.widget_event.setVisible(show_event)
        self.widget_scatter.select.setVisible(show_event)  # point in scatter
        self.widget_poly.setVisible(show_poly)
        self.widget_settings.setVisible(show_settings)

        if show_event:
            # update event plot (maybe axes changed)
            self.on_event_scatter_update()

        for b in toblock:
            b.blockSignals(False)

        if not show_poly:
            self.on_poly_done()
        # set size
        self.update()
        ws = self.sizeHint()
        mdiwin = self.parent()
        geom = mdiwin.geometry()
        geom.setWidth(ws.width())
        geom.setHeight(ws.height())
        mdiwin.setGeometry(geom)
        mdiwin.adjustSize()
        mdiwin.update()

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def plot(self):
        """Update the plot using the current state of the UI"""
        if self.rtdc_ds is not None:
            plot = self.__getstate__()["plot"]
            downsample = plot["downsampling"] * plot["downsampling value"]

            self.widget_scatter.plot_data(rtdc_ds=self.rtdc_ds,
                                          slot=self.slot,
                                          downsample=downsample,
                                          xax=plot["axis x"],
                                          yax=plot["axis y"],
                                          xscale=plot["scale x"],
                                          yscale=plot["scale y"],
                                          isoelastics=plot["isoelastics"])
            # make sure the correct plot items are visible
            # (e.g. scatter select)
            self.on_tool()
            # update polygon filter axis names
            self.label_poly_x.setText(
                dclab.dfn.get_feature_label(plot["axis x"]))
            self.label_poly_y.setText(
                dclab.dfn.get_feature_label(plot["axis y"]))
            self.show_statistics()

    @show_wait_cursor
    @QtCore.pyqtSlot(int)
    def show_event(self, event):
        """Display the event data (image, contour, trace)

        Parameters
        ----------
        event: int
            Event index of the dataset; indices start at 0
            If set to None, the index from `self.spinBox_event`
            will be used.
        """
        # dataset
        ds = self.rtdc_ds
        event_count = ds.config["experiment"]["event count"]
        if event_count == 0:
            # nothing to do
            return
        # Update spin box data
        self.spinBox_event.blockSignals(True)
        self.spinBox_event.setValue(event + 1)
        self.spinBox_event.blockSignals(False)

        # Update selection point in scatter plot
        self.widget_scatter.setSelection(event)
        if self.tabWidget_event.currentIndex() == 0:
            # update image
            state = self.__getstate__()
            if "image" in ds:
                cellimg, imkw = self.get_event_image(ds, event)
                self.imageView_image.setImage(cellimg, **imkw)
                self.groupBox_image.show()
            else:
                self.groupBox_image.hide()
            if "trace" in ds:
                # remove legend items
                for item in reversed(self.legend_trace.items):
                    self.legend_trace.removeItem(item[1].text)
                self.legend_trace.setVisible(state["event"]["trace legend"])
                # get slot from identifier
                # time axis
                flsamples = ds.config["fluorescence"]["samples per event"]
                flrate = ds.config["fluorescence"]["sample rate"]
                fltime = np.arange(flsamples) / flrate * 1e6
                # temporal range (min, max, fl-peak-maximum)
                range_t = [fltime[0], fltime[-1], 0]
                # fluorescence intensity
                range_fl = [0, 0]
                for key in dclab.dfn.FLUOR_TRACES:
                    if key.count("raw") and not state["event"]["trace raw"]:
                        # hide raw trace data if user decided so
                        show = False
                    else:
                        show = True
                    flid = key.split("_")[0]
                    if (key in ds["trace"] and show):
                        # show the trace information
                        tracey = ds["trace"][key][event]  # trace data
                        range_fl[0] = min(range_fl[0], tracey.min())
                        range_fl[1] = max(range_fl[1], tracey.max())
                        self.trace_plots[key].setData(fltime, tracey)
                        self.trace_plots[key].show()
                        if state["event"]["trace zoom"]:
                            flpos = ds["{}_pos".format(flid)][event]
                            flwidth = ds["{}_width".format(flid)][event]
                            flmax = ds["{}_max".format(flid)][event]
                            # use the peak maximum to decide which range to use
                            if flmax > range_t[2]:
                                range_t[0] = flpos - 1.5 * flwidth
                                range_t[1] = flpos + 1.5 * flwidth
                                range_t[2] = flmax
                        # set legend name
                        ln = "{} {}".format(self.slot.fl_name_dict[
                            "FL-{}".format(key[2])], key[4:])
                        self.legend_trace.addItem(self.trace_plots[key], ln)
                        self.legend_trace.update()
                    else:
                        self.trace_plots[key].hide()
                self.graphicsView_trace.setXRange(*range_t[:2], padding=0)
                if range_fl[0] != range_fl[1]:
                    self.graphicsView_trace.setYRange(*range_fl, padding=.01)
                self.graphicsView_trace.setLimits(xMin=0, xMax=fltime[-1])
                self.groupBox_trace.show()
            else:
                self.groupBox_trace.hide()
        else:
            # only use computed features (speed)
            fcands = ds.features_loaded
            feats = [f for f in fcands if f in ds.features_scalar]
            lf = sorted([(dclab.dfn.get_feature_label(f), f) for f in feats])
            keys = []
            vals = []
            for lii, fii in lf:
                keys.append(lii)
                val = ds[fii][event]
                if fii in idiom.INTEGER_FEATURES:
                    val = int(np.round(val))
                vals.append(val)
            self.tableWidget_feats.set_key_vals(keys, vals)

    @show_wait_cursor
    def show_rtdc(self, rtdc_ds, slot):
        """Display an RT-DC measurement given by `path` and `filters`"""
        # Create a hierarchy child so that the user can browse
        # comfortably through the data without seeing hidden events.
        self.rtdc_ds = dclab.new_dataset(rtdc_ds)
        event_count = self.rtdc_ds.config["experiment"]["event count"]
        if event_count == 0:
            self.widget_scatter.hide()
            self.widget_tool.setEnabled(False)
            self.label_noevents.show()
            self.on_tool(collapse=True)
            return
        else:
            # make things visible
            self.label_noevents.hide()
            self.label_howto.hide()
            self.widget_scatter.show()
            self.widget_tool.setEnabled(True)
        # get the state
        state = self.__getstate__()
        plot = state["plot"]
        # remove event state (ill-defined for different datasets)
        state.pop("event")

        self.slot = slot
        # default features (plot axes)
        if plot["axis x"] is None:
            plot["axis x"] = "area_um"
        if plot["axis y"] is None:
            plot["axis y"] = "deform"
        # check whether axes exist in ds and change them to defaults
        # if necessary
        ds_features = self.rtdc_ds.features_scalar
        if plot["axis x"] not in ds_features:
            for feat in dclab.dfn.scalar_feature_names:
                if feat in ds_features:
                    plot["axis x"] = feat
                    break
        if plot["axis y"] not in ds_features:
            for feat in dclab.dfn.scalar_feature_names:
                if feat in ds_features:
                    plot["axis y"] = feat
                    if feat != plot["axis y"]:
                        # If there is only one feature, at least we
                        # have set the state to a reasonable value.
                        break
        # set control ranges
        self.spinBox_event.blockSignals(True)
        self.spinBox_event.setMaximum(event_count)
        self.spinBox_event.setToolTip("total: {}".format(event_count))
        self.spinBox_event.setValue(1)
        self.spinBox_event.blockSignals(False)

        # set quick view state
        self.__setstate__(state)
        # scatter plot
        self.plot()
        # select first event in event viewer (also updates selection point)
        self.show_event(0)
        # this only updates the size of the tools (because there is no
        # sender)
        self.on_tool()

    def show_statistics(self):
        h, v = self.get_statistics()
        if h is not None:
            self.tableWidget_stat.set_key_vals(keys=h, vals=v)

    def update_feature_choices(self):
        """Updates the axes comboboxes choices

        This is used e.g. when emodulus becomes available
        """
        if self.rtdc_ds is not None:
            # axes combobox choices
            ds_feats = self.rtdc_ds.features_scalar
            ds_labels = [dclab.dfn.get_feature_label(f) for f in ds_feats]
            ds_fl = sorted(zip(ds_labels, ds_feats))
            for cb in [self.comboBox_x, self.comboBox_y]:
                fcur = cb.currentData()
                blocked = cb.signalsBlocked()  # remember block state
                cb.blockSignals(True)
                # set features
                cb.clear()
                for label, feat in ds_fl:
                    if feat in ds_feats:
                        cb.addItem(label, feat)
                idcur = cb.findData(fcur)
                if idcur >= 0:
                    cb.setCurrentIndex(idcur)
                cb.blockSignals(blocked)

    def update_polygon_panel(self):
        """Update polygon filter combobox etc."""
        pfts = dclab.PolygonFilter.instances
        if pfts:
            self.comboBox_poly.blockSignals(True)
            self.comboBox_poly.clear()
            self.comboBox_poly.addItem("Choose...", None)
            for pf in pfts:
                self.comboBox_poly.addItem(pf.name, pf.unique_id)
            self.comboBox_poly.blockSignals(False)
            self.comboBox_poly.setEnabled(True)
        else:
            self.comboBox_poly.setEnabled(False)
        self.groupBox_poly.setEnabled(False)



import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets


class AnalysisView(QtWidgets.QWidget):
    filter_changed = QtCore.pyqtSignal(dict)
    plot_changed = QtCore.pyqtSignal(dict)
    slot_changed = QtCore.pyqtSignal(dict)
    pipeline_changed = QtCore.pyqtSignal(dict)

    def __init__(self, *args, **kwargs):
        QtWidgets.QWidget.__init__(self)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.analysis", "ana_view.ui")
        uic.loadUi(path_ui, self)
        self.setWindowTitle("Analysis View")
        self.setMinimumSize(self.sizeHint())
        # Signals
        self.widget_filter.filter_changed.connect(self.filter_changed)
        self.widget_filter.pipeline_changed.connect(self.pipeline_changed)
        self.widget_plot.plot_changed.connect(self.plot_changed)
        self.widget_plot.pipeline_changed.connect(self.pipeline_changed)
        self.widget_slot.slot_changed.connect(self.slot_changed)
        self.widget_slot.pipeline_changed.connect(self.pipeline_changed)

    def adopt_pipeline(self, pipeline_state):
        self.widget_meta.pipeline_state = pipeline_state
        # widget_plot, widget_slot, and widget_filter know the pipeline
        self.update_content()

    def set_pipeline(self, pipeline):
        self.widget_filter.set_pipeline(pipeline)
        self.widget_plot.set_pipeline(pipeline)
        self.widget_slot.set_pipeline(pipeline)
        self.adopt_pipeline(pipeline.__getstate__())

    def update_content(self):
        self.widget_filter.update_content()
        self.widget_meta.update_content()
        self.widget_plot.update_content()
        self.widget_slot.update_content()



import functools
import hashlib
import io
import json
import os
import pathlib
import shutil
import tempfile
import zipfile

import dclab

from .pipeline import Dataslot, Filter, Pipeline, Plot
from ._version import version


class DataFileNotFoundError(BaseException):
    def __init__(self, missing_paths, *args, **kwargs):
        self.missing_paths = missing_paths
        super(DataFileNotFoundError, self).__init__(*args, **kwargs)


class PathlibJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, pathlib.Path):
            return {"__type__": "path",
                    "__data__": o.as_posix()
                    }
        # Let the base class default method raise the TypeError
        return super(PathlibJSONEncoder, self).default(o)


class PathlibJSONDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        super(PathlibJSONDecoder, self).__init__(object_hook=self.object_hook,
                                                 *args, **kwargs)

    def object_hook(self, obj):
        if "__type__" in obj and obj["__type__"] == "path":
            return pathlib.Path(obj["__data__"])
        return obj


def export_filters(path, pipeline, filt_ids=None):
    """Export filters of a pipeline to a JSON file"""
    if filt_ids is None:
        filt_ids = pipeline.filter_ids
    # First, get all relevant polygon filters
    polyids = set([])
    for filt in pipeline.filters:
        if filt.identifier in filt_ids:
            polyids |= set(filt.polylist)
    poly_states = []
    for pid in sorted(polyids):
        pf = dclab.PolygonFilter.get_instance_from_id(pid)
        poly_states.append(pf.__getstate__())
    # Then, get the remaining filter settings
    filt_states = []
    for filt in pipeline.filters:
        if filt.identifier in filt_ids:
            filt_states.append(filt.__getstate__())
    state = {"polygon filters": poly_states,
             "filters": filt_states}
    dump = json.dumps(state, sort_keys=True, indent=2)
    pathlib.Path(path).write_text(dump)


def import_filters(path, pipeline, strict=False):
    """Load filters from a JSON file into a pipeline

    Parameters
    ----------
    path: pathlib.Path or str
        Path to the filter file
    pipeline: shapeout2.pipeline.Pipeline
        Analysis pipeline to import filters to
    strict: bool
        If False (default), new filter identifiers are created
        for each filter. If True, the given filter identifiers
        will be used - Raises a ValueError if the filter exists.
        For loading filters into a pipeline, this should be False.
        For loading session filters, this should be True.
        Has no effect when `path` is a .poly file.
    """
    if isinstance(path, io.IOBase):
        import_filter_set(path, pipeline, strict)
    else:
        path = pathlib.Path(path)
        if path.suffix == ".poly":
            # add a new polygon filter
            dclab.PolygonFilter.import_all(path)
        elif path.suffix == ".sof":
            import_filter_set(path, pipeline, strict)
        else:
            raise ValueError("Unrecognized file extension "
                             + "'{}' for filters.".format(path.suffix))


def import_filter_set(path, pipeline, strict=False):
    """Import a filter set

    See :func:`import_filters`
    """
    if isinstance(path, io.IOBase):
        dump = path.read()
    else:
        dump = path.read_text()
    dump_state = json.loads(dump)
    # add polygon filters from file
    if not strict:
        pf_dict = {}  # maps old to new identifiers
    for pstate in dump_state["polygon filters"]:
        pf = dclab.PolygonFilter(axes=(pstate["axis x"], pstate["axis y"]),
                                 points=pstate["points"])
        pid = pstate["identifier"]
        if not strict:
            # keep track of old and new identifiers
            pf_dict[pid] = pf.unique_id
        if pid != pf.unique_id:
            if strict:
                if dclab.PolygonFilter.unique_id_exists(pid):
                    raise ValueError("PolygonFilter with unique_id "
                                     + "{} already exists!".format(pid))
                else:
                    # change the unique_id to that of the original filter
                    pf._set_unique_id(pid)
            else:
                # use the unique_id of the newly-created filter
                pstate["identifier"] = pf.unique_id
        pf.__setstate__(pstate)
    # add a new filter set
    for state in dump_state["filters"]:
        if strict:
            filt = Filter(identifier=state["identifier"])
        else:
            filt = Filter()
            state["identifier"] = filt.identifier
            # transform original polygon filter ids
            newpids = [pf_dict[pid] for pid in state["polygon filters"]]
            state["polygon filters"] = newpids
        filt.__setstate__(state)
        pipeline.add_filter(filt=filt)


@functools.lru_cache(maxsize=1000)
def hash_file_partially(path, size=524288):
    """Hash parts of a file for basic identification

    By default, the first and final 512kB are hashed.
    """
    fsize = path.stat().st_size
    size = min(size, fsize)
    with path.open("rb") as fd:
        head = fd.read(size)
        fd.seek(fsize-size)
        tail = fd.read(size)
        hexhash = hashlib.md5(head + tail).hexdigest()
    return hexhash


def save_session(path, pipeline):
    """Save an entire pipeline session"""
    path = pathlib.Path(path)
    tempdir = pathlib.Path(tempfile.mkdtemp(prefix="ShapeOut2-session-save_"))
    # filters
    export_filters(tempdir / "filters.sof", pipeline)
    # slots
    for ii, slot in enumerate(pipeline.slots):
        sdump = json.dumps(slot.__getstate__(),
                           cls=PathlibJSONEncoder,
                           sort_keys=True,
                           indent=2)
        (tempdir / "slot_{}.json".format(ii)).write_text(sdump)
    # plots
    for jj, plot in enumerate(pipeline.plots):
        pdump = json.dumps(plot.__getstate__(),
                           sort_keys=True,
                           indent=2)
        (tempdir / "plot_{}.json".format(jj)).write_text(pdump)
    # pipeline block matrix
    mdump = json.dumps(pipeline.element_states,
                       sort_keys=True,
                       indent=2)
    (tempdir / "matrix.json").write_text(mdump)
    # additional information
    search_paths = {}
    dataset_hashes = {}
    dataset_formats = {}
    for slot in pipeline.slots:
        # format (hdf5 or dcor)
        dataset_formats[slot.identifier] = slot.format
        if slot.format == "hdf5":
            # search path
            try:
                rel = os.path.relpath(slot.path.parent, path.parent)
            except (OSError, ValueError):
                rel = "."
            search_paths[slot.identifier] = rel
            # file hash
            hash_size = 48128
            dataset_hashes[slot.identifier] = {
                # these are keyword arguments to find_file
                "partial_hash": hash_file_partially(slot.path, size=hash_size),
                "size_read": hash_size
            }
    remarks = {"search paths": search_paths,
               "file hashes": dataset_hashes,
               "formats": dataset_formats,
               "version": version,
               }
    rdump = json.dumps(remarks,
                       cls=PathlibJSONEncoder,
                       sort_keys=True,
                       indent=2)
    (tempdir / "remarks.json").write_text(rdump)
    # zip everything
    with zipfile.ZipFile(path, mode='w') as arc:
        for pp in tempdir.rglob("*"):
            if pp.is_file():
                arc.write(pp, pp.relative_to(tempdir))
    # cleanup
    shutil.rmtree(str(tempdir), ignore_errors=True)


def clear_session(pipeline=None):
    """Clear the entire analysis pipeline"""
    if pipeline is not None:
        # reset the pipeline
        pipeline.reset()
    # Close all file handles
    for slot_id in list(Dataslot._instances.keys()):
        Dataslot.remove_slot(slot_id)
    # remove any existing filters, plots, or slots and reset their counters
    for cls in [Dataslot, Filter, Plot]:
        cls._instance_counter = 0
        cls._instances = {}
    # remove polygon filters
    dclab.PolygonFilter.clear_all_filters()


def find_file(original_path, search_paths, partial_hash, size_read):
    """Find a file

    Parameters
    ----------
    original_path: pathlib.Path
        The original path to the file
    search_paths: list
        Directories or possible candidates for the file. In
        directories, only files with the same name are searched.
    partial_hash: str
        Hash of the file, see :func:`hash_file_partially`

    Returns a pathlib.Path object on success, False otherwise.
    """
    search_paths = [pathlib.Path(sp) for sp in search_paths]
    if original_path.exists():
        # we boldly assume that the hash matches
        path = original_path
    else:
        for pp in search_paths:
            if pp.is_dir():
                # look in the directory
                newp = pp / original_path.name
                if newp.exists():
                    newh = hash_file_partially(newp, size=size_read)
                    if newh == partial_hash:
                        path = newp
                        break
            elif pp.exists():
                # maybe this is the file?
                newh = hash_file_partially(pp, size=size_read)
                if newh == partial_hash:
                    path = pp
                    break
        else:
            path = False
    return path


def open_session(path, pipeline=None, search_paths=[]):
    """Load a session (optionally overriding an existing pipeline)

    Parameters
    ----------
    path: pathlib.Path or str
        Path to the session file
    pipeline: shapeout2.pipeline.Pipeline or None
        If a pipeline is given, it is reset before loading anything
    search_paths: list
        Paths to search for missing measurements; entries may be
        directories or .rtdc files
    """
    path = pathlib.Path(path)
    if pipeline is None:
        pipeline = Pipeline()
    clear_session(pipeline)
    # read data directly from the zip file
    with zipfile.ZipFile(path, mode='r') as arc:
        # remarks
        remarks = json.loads(arc.read("remarks.json"))
        # determine dataset paths from slot states
        slotnames = sorted(
            [n for n in arc.namelist() if n.startswith("slot_")],
            key=lambda x: int(x[5:-5]))  # by index
        slot_states = []
        missing_paths = []
        for sn in slotnames:
            sstate = json.loads(arc.read(sn), cls=PathlibJSONDecoder)
            slot_id = sstate["identifier"]
            # "formats" was added in 2.1.0 (when dcor format was added)
            ishdf5 = ("formats" not in remarks
                      or remarks["formats"][slot_id] == "hdf5")
            if ishdf5:
                # also search relative paths
                search_ap = path.parent / remarks["search paths"][slot_id]
                newpath = find_file(original_path=sstate["path"],
                                    search_paths=search_paths + [search_ap],
                                    **remarks["file hashes"][slot_id])
                if newpath:
                    sstate["path"] = newpath
                    slot_states.append(sstate)
                else:
                    missing_paths.append(sstate["path"])
            else:
                slot_states.append(sstate)
        # raise an exception if data files are missing
        if missing_paths:
            # Which files are missing is stored as a property in the exception.
            # By catching this exception, a GUI can request the user to select
            # a search directory and try again.
            raise DataFileNotFoundError(
                missing_paths,
                "Some files are missing! You can access them via the "
                + "`missing_paths` property of this exception.")
        # load filters
        import_filters(arc.open("filters.sof"), pipeline, strict=True)
        # load slots
        for sstate in slot_states:
            slot = Dataslot(identifier=sstate["identifier"],
                            path=sstate["path"])
            slot.__setstate__(sstate)
            pipeline.add_slot(slot)
        # load plots
        plotnames = sorted(
            [n for n in arc.namelist() if n.startswith("plot_")],
            key=lambda x: int(x[5:-5]))  # by index
        for pn in plotnames:
            pstate = json.loads(arc.read(pn))
            plot = Plot(identifier=pstate["identifier"])
            plot.__setstate__(pstate)
            pipeline.add_plot(plot)
        # load element states
        estates = json.loads(arc.read("matrix.json"))
        pipeline.element_states = estates
    return pipeline



import pathlib
import pkg_resources

from PyQt5 import uic, QtWidgets

import dclab

from ... import session
from ...util import get_valid_filename


class ExportFilter(QtWidgets.QDialog):
    def __init__(self, parent, pipeline, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.export", "e2filter.ui")
        uic.loadUi(path_ui, self)
        # set pipeline
        self.pipeline = pipeline

    @property
    def file_format(self):
        if self.radioButton_poly.isChecked():
            return "poly"
        else:
            return "sof"

    @property
    def file_mode(self):
        if self.radioButton_single.isChecked():
            return "single"
        else:
            return "multiple"

    def done(self, r):
        if r:
            self.export_filters()
        super(ExportFilter, self).done(r)

    def export_filters(self):
        """Export filters"""
        if self.file_mode == "single":
            # all in one file
            if self.file_format == "poly":
                cap = "Polygon filter files (*.poly)"
                path, _ = QtWidgets.QFileDialog.getSaveFileName(
                    self, 'Save polygon filters', '', cap)
                if not path.endswith(".poly"):
                    path += ".poly"
                dclab.PolygonFilter.save_all(path)
            else:
                cap = "Shape-Out 2 filter files (*.sof)"
                path, _ = QtWidgets.QFileDialog.getSaveFileName(
                    self, 'Save polygon filters', '', cap)
                if not path.endswith(".sof"):
                    path += ".sof"
                session.export_filters(path=path,
                                       pipeline=self.pipeline)
        else:
            # one file per filter
            path = QtWidgets.QFileDialog.getExistingDirectory(self,
                                                              'Output Folder')
            if self.file_format == "poly":
                for pf in dclab.PolygonFilter.instances:
                    fn = "SO2-PolygonFilter_{}_{}.poly".format(pf.unique_id,
                                                               pf.name)
                    # remove bad characters from file name
                    fn = get_valid_filename(fn)
                    out = pathlib.Path(path) / fn
                    pf.save(out)
            else:
                for filt_index, filt_id in enumerate(self.pipeline.filter_ids):
                    filt = self.pipeline.filters[filt_index]
                    fn = "SO2-Filter_{}_{}.sof".format(filt_index, filt.name)
                    # remove bad characters from file name
                    fn = get_valid_filename(fn)
                    out = pathlib.Path(path) / fn
                    session.export_filters(path=out,
                                           pipeline=self.pipeline,
                                           filt_ids=[filt_id])



import pkg_resources
import requests
import traceback as tb
import urllib.parse
import webbrowser

import dclab
from PyQt5 import uic, QtCore, QtGui, QtWidgets

from ..widgets import show_wait_cursor, run_async


class DCORLoader(QtWidgets.QDialog):
    search_finished = QtCore.pyqtSignal(int, list, list, object)
    search_item_retrieved = QtCore.pyqtSignal(int, int, dict, dict, str)

    def __init__(self, parent, *args, **kwargs):
        """Search and load DCOR data"""
        QtWidgets.QWidget.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "shapeout2.gui.dcor", "dcor.ui")
        uic.loadUi(path_ui, self)

        self.main_ui = parent
        self.search_results = []
        self.num_searches = 0

        # Update UI
        self.settings = QtCore.QSettings()

        # tool button
        self.pushButton_search.clicked.connect(self.on_search)
        searchicon = QtGui.QIcon.fromTheme("search")
        self.pushButton_search.setIcon(searchicon)
        self.pushButton_search.setDefault(True)
        self.buttonBox.buttons()[1].setDefault(False)
        self.buttonBox.buttons()[0].setDefault(False)
        self.lineEdit_search.setFocus()

        # search signals
        self.search_finished.connect(self.on_search_finished)
        self.search_item_retrieved.connect(self.on_search_add_result)

        # button signals
        btn_close = self.buttonBox.button(QtGui.QDialogButtonBox.Close)
        btn_close.clicked.connect(self.on_close)
        btn_close.setToolTip("Close this window")
        closeicon = QtGui.QIcon.fromTheme("dialog-close")
        btn_close.setIcon(closeicon)
        btn_open = self.buttonBox.button(QtGui.QDialogButtonBox.Apply)
        btn_open.clicked.connect(self.on_open)
        btn_open.setToolTip("Add selected resources to current session")
        btn_open.setText("Add to session")
        plusicon = QtGui.QIcon.fromTheme("list-add")
        btn_open.setIcon(plusicon)
        btn_help = self.buttonBox.button(QtGui.QDialogButtonBox.Help)
        btn_help.clicked.connect(self.on_help)
        btn_help.setToolTip("View DCOR Quick Guide online")
        helpicon = QtGui.QIcon.fromTheme("documentinfo")
        btn_help.setIcon(helpicon)

    def get_api_base_url(self):
        """Return the API url in the form https://dcor.mpl.mpg.de/api/3"""
        server = self.settings.value("dcor/servers", ["dcor.mpl.mpg.de"])[0]
        use_ssl = bool(int(self.settings.value("dcor/use ssl", 1)))
        # ready API
        http = "https" if use_ssl else "http"
        base = "{}://{}/api/3".format(http, server)
        return base

    def get_api_headers(self):
        """Return the API headers (Authorization with API key)"""
        api_key = self.settings.value("dcor/api key", "")
        # Add this API Key to the known API keys (dclab)
        dclab.rtdc_dataset.fmt_dcor.APIHandler.add_api_key(api_key)

        if api_key:
            api_headers = {"Authorization": api_key}
        else:
            api_headers = {}
        return api_headers

    @QtCore.pyqtSlot()
    def on_close(self):
        """Close window"""
        self.close()

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def on_open(self):
        """Add selected resources to the current session"""
        for ii in range(self.listWidget.count()):
            item = self.listWidget.item(ii)
            if item.isSelected():
                self.main_ui.add_dataslot(
                    paths=[self.search_results[ii]], is_dcor=True)

    @QtCore.pyqtSlot()
    def on_help(self):
        """Show Shape-Out 2 docs"""
        webbrowser.open(
            "https://shapeout2.readthedocs.io/en/stable/sec_qg_dcor.html")

    @run_async  # comment-out for debugging
    @show_wait_cursor
    @QtCore.pyqtSlot()
    def on_search(self):
        """Trigger a search given the current search settings

        Notes
        -----
        This function is run in a background thread to not block
        the user interface. While this function is running, the
        user may start a new search. For each search, the counter
        ``self.num_searches`` is incremented. After a search is
        complete, the current search id is checked against
        ``self.num_searches`` and only if they match are the
        results displayed in the UI.

        See Also
        --------
        on_search_add_result: called for every search result
        on_search_finished: called when search finishes
        """
        self.num_searches += 1
        this_search_id = self.num_searches

        api_base_url = self.get_api_base_url()
        api_headers = self.get_api_headers()

        # search string
        if self.comboBox_search.currentIndex() == 1:
            stype = "dataset"
        else:
            stype = "free"
        search_string = urllib.parse.quote(self.lineEdit_search.text())

        # perform search
        try:
            # Results are handled individually via the `search_item_retrieved`
            # event.
            results, failed = self.perform_search(
                search_string=search_string,
                search_type=stype,
                search_id=this_search_id,
                api_base_url=api_base_url,
                api_headers=api_headers,
            )
        except BaseException:
            results = []
            failed = []
            error = tb.format_exc(limit=2, chain=False)
        else:
            error = False
        self.search_finished.emit(this_search_id, results, failed, error)

    @QtCore.pyqtSlot(int, int, dict, dict, str)
    def on_search_add_result(self, search_id, result_index, dataset, resource,
                             api_base_url):
        """Add new item to ``self.listWidget`` and ``self.search_results``"""
        if search_id == self.num_searches:
            if result_index == 0:
                self.listWidget.clear()
                self.search_results.clear()
            name = "{}: {} <{}@{}>".format(
                dataset["title"],
                resource["name"],
                dataset["name"],
                dataset["organization"]["name"],
            )
            self.listWidget.addItem(name)
            ru = api_base_url + "/action/dcserv?id={}".format(resource["id"])
            self.search_results.append(ru)

    @QtCore.pyqtSlot(int, list, list, object)
    def on_search_finished(self, search_id, results, failed, error):
        """Finalize search

        Finalization includes:

        - clear ``self.listWidget`` if the search had no results
        - display messages about datasets that were not displayed
        - display error messages
        """
        if search_id != self.num_searches:
            # user triggered next search
            return
        elif not results:
            # no results for this search
            self.listWidget.clear()

        if failed:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setText("Search found invalid data: {}".format(failed))
            msg.setWindowTitle("Dataset validation")
            msg.exec_()

        if error:
            msg = QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setText(error)
            msg.setWindowTitle("DCOR access error!")
            msg.exec_()

    def perform_search(self, search_string, search_type, search_id,
                       api_base_url, api_headers):
        """Perform search

        Parameters
        ----------
        search_string: str
            Search string (already parsed using urllib.parse.quote)
        search_type: str
            "free": free text search
            or "dataset": resource/package name/id
        search_id: int
            Search identifier (must match `self.num_searches` for
            the search to continue)
        api_base_url: str
            Everything up until "https://server.example.org/api/3"
        api_headers: dict
            Headers for the request
        """
        if search_type == "free":
            url = api_base_url + "/action/package_search?q={}".format(
                search_string)
            # limit to 20 rows
            url += "&rows=20"
            # include private data
            url += "&include_private=True"
            urls = [url]
        else:
            urls = [
                api_base_url + "/action/package_show?id={}".format(
                    search_string),
                api_base_url + "/action/resource_show?id={}".format(
                    search_string),
            ]

        pkg_res = []
        for url in urls:
            req = requests.get(url, headers=api_headers)
            if not req.ok:
                continue
            resp = req.json()["result"]
            if "count" in resp and "results" in resp:
                # free search
                for pkg in resp["results"]:
                    if "resources" in pkg:
                        for res in pkg["resources"]:
                            pkg_res.append([pkg, res])
            elif "resources" in resp:
                # package show
                for res in resp["resources"]:
                    pkg_res.append([resp, res])
                break  # no additional search for resource
            else:
                # resource show
                purl = api_base_url + "/action/package_show?id={}".format(
                    resp["package_id"])
                pkg = requests.get(purl, headers=api_headers).json()["result"]
                pkg_res.append([pkg, resp])
                break  # no additional search for package

        results = []
        failed = []
        for ii, (pkg, res) in enumerate(pkg_res):
            if res["mimetype"] not in ["RT-DC"]:
                continue  # only use RT-DC data
            if search_id != self.num_searches:
                break  # stop doing things in the background
            # check availability of each resource
            c = api_base_url + "/action/dcserv?id={}&query=valid".format(
                res["id"])
            req = requests.get(c, headers=api_headers)
            if req.ok and req.json()["result"]:  # only use valid data
                self.search_item_retrieved.emit(search_id, ii, pkg, res,
                                                api_base_url)
                results.append(res)
            else:
                failed.append("{}: {}".format(res["id"], req.reason))
        return results, failed
